; SPDX license id: GPL-2.0-or-later
;
; AVR Sound Source
; Version: 0.5b
; (C) 2023 Shamatic
;
; DSS/Covox/Stereo-on-1 auto-switching emulator
; written in glorious AVR Assembly for Atmega*8

; NOTE: Set the DFIFO_DIVIDER value for 7KHz DSS output:
; F_CPU/7000/128-1
; For example:
; 	25000000Hz 27
; 	22579200Hz 24
; 	20000000Hz 21
; 	16000000Hz 17
; 22.5792 MHz crystal is recommended for sound quality

.equ DFIFO_DIVIDER, 24

.equ D_WAIT, 0x1F		;timeout for data waitage, 7*clocks
.equ D_FALLBACKH, 0x03	;timeout to fall back to mono mode, 4*FF clocks

.equ B_MAX, 0x0F		;DSS buffer size

.equ P_STROBE, 0		;LPT pin 1 on PB
.equ P_SELECT, 0		;LPT pin 17 on PC
.equ P_ACK, 3			;LPT pin 10 on PC
.equ S_LED, 5			;Stereo/DSS led on PB

.equ PINB, 0x03
.equ PORTB, 0x05
.equ PINC, 0x06
.equ PORTC, 0x08
.equ PIND, 0x09
.equ PCICR, 0x68
.equ OCR1AL, 0x88
.equ OCR1BL, 0x8A
.equ TCCR2A, 0xB0
.equ SPL, 0x3D

.section .text

; --- INTERRUPT VECTORS ---
.org 0x00				;Reset, to init code
	rjmp Main

.org 0x0C				;PCINT0, stereo input strobe
	in r24, PIND
	rjmp Stereo_Input

.org 0x10				;PCINT1, dss input strobe
	in r24, PIND
	sbic PINC, P_SELECT	;if dss select is downward
	rjmp dReturn
	rjmp Dss_Input

; --- DSS OUTPUT TIMER ---
.org 0x1C ;TIMER2 COMPA
	cp   XL, YL			;if buffer pointers not equal
	brne DO_Output

	sbic PINC, P_ACK	;if ack is on
	rjmp DO_AckTurnOff

	sts  TCCR2A, r1		;disable dss output timer
	rjmp dReturn

DO_AckTurnOff:
	cbi  PORTC, P_ACK	;turn off ack
	sts  PCICR, r18		;enable DSS input interrupt
DO_Output:
	ld   r24, X+		;load from buffer and step
	sts  OCR1AL, r24	;left channel out
	sts  OCR1BL, r24	;right channel out

	andi XL, B_MAX		;buffer limit
	rjmp dReturn

; --- DSS INPUT TRIGGER ---
Dss_Input:
	st   Y+, r24		;save to buffer and step
	andi YL, B_MAX		;buffer limit

	cp   XL, YL			;if buffer pointers not equal
	brne DI_Setregs

	sbi  PORTC, P_ACK	;set ack
	sts  PCICR, r1		;disable DSS input interrupt
	rjmp dReturn

DI_Setregs:
	sts  PCICR, r18		;dss strobe only
	sts  TCCR2A, r18	;enable dss output timer
	rjmp dReturn

; --- STROBE INPUT TRIGGER ---
Stereo_Input:
	sbic PINB, P_STROBE	;downward only
	rjmp sReturn

	ldi  r23, D_WAIT
SI_Delay:				;wait for next data
	in   r25, PIND
	cp   r24, r25
	brne SI_RightAndOut
	dec  r23
	brne SI_Delay

SI_RightAndOut:
	sbic PINB, P_STROBE	;are we still low?
	rjmp SI_ReverseChannel

	in   r25, PIND		;read again
	sts  OCR1AL, r24	;left channel out
	sts  OCR1BL, r25	;right channel out
	rjmp sReturn

SI_ReverseChannel:
	in   r25, PIND		;read again
	sts  OCR1AL, r25	;left channel out
	sts  OCR1BL, r24	;right channel out
	rjmp sReturn

; --- POWER ON SETUP ---
Main:
	; Power saving
	ldi  r24, 0b10100111	;everything is off, except timer 0 & 2
	sts  0x64, r24			;PRR

	; DSS buffer pointers
	eor  XL, XL				;read position low
	ldi  XH, 0x01			;read position high
	eor  YL, YL				;write position low
	ldi  YH, 0x01			;write position high

	; registers for setting things
	eor  r1, r1				;zero
	ldi  r18, 0b00000010	;PCICR: PCIE1 | TCCR2A: WGM21 | TIMSK2: OCIE2A
	ldi  r19, 0b00000011 	;PCICR: PCIE1, PCIE0
	ldi  r20, 0b00000001 	;PCICR: PCIE0
	ldi  r21, 0xFF			;stack pointer low

	; we still need to set these, because we're using interrupts
	ldi  r24, 0x01			;stack pointer high
	out  0x3E, r24			;SPH
	out  SPL, r21

	; PWM output setup
	ldi  r24, 0b11110001 	;-- COM1A1, COM1A0, COM1B1, COM1B0, WGM10
	sts  0x80, r24			;TCCR1A
	ldi  r24, 0b00001001	;-- CS10, WGM12 (fast pwm)
	sts  0x81, r24			;TCCR1B
	ldi  r24, 0b00100110	;-- PB5, PB2, PB1
	out  0x04, r24			;DDRB

	; DSS timer setup
	sts  0x70, r18			;TIMSK2
	ldi  r24, 0b00000101	;-- CS22, CS20
	sts  0xB1, r24			;TCCR2B
	ldi  r24, DFIFO_DIVIDER
	sts  0xB3, r24			;OCR2A
	sbi  0x07, P_ACK		;DDRC

	; PIN change interrupts
	ldi  r24, (1<<P_STROBE)
	sts  0x6B, r24			;PCMSK0
	ldi  r24, (1<<P_SELECT)
	sts  0x6C, r24			;PCMSK1
	sts  PCICR, r19			;PCICR

	; Input pull-ups (might be required for some weak LPT ports)
	;out 0x0B, r24			;PORTD pull-up

	sei						;enable interrupts

; --- VARIOUS RUNNING STATES ---
	in   r24, PIND
mLoop:				;MONO COVOX LOOP, non-scratchy edition
	in   r25, PIND
	cp   r24, r25
	brne mDataOut
	rjmp mLoop

mDataOut:
	nop						;tactical data-stabilizing delay
	nop
	in   r24, PIND
	sts  OCR1AL, r24
	sts  OCR1BL, r24
	rjmp mLoop

sReturn:			;STEREO RETURN
	sts  PCICR, r20			;stereo strobe only
dReturn:			;DSS RETURN
	out  SPL, r21			;counteract interrupt behavior
	sei
	sbi  PORTB, S_LED

	ldi  r25, D_FALLBACKH
	ldi  r24, 0xFF
modeWait:					;timeout to fall back to mono
	sbiw r24, 1
	brne modeWait

	sts  PCICR, r19			;all pc interrupts enable
	cbi  PORTB, S_LED
	rjmp mLoop
.end
